import numpy as np
from scipy.optimize import linear_sum_assignment
from ._base_metric import _BaseMetric
from .. import _timing
from collections import defaultdict
from .. import utils


class IDEucl(_BaseMetric):
    """Class which implements the ID metrics"""

    @staticmethod
    def get_default_config():
        """Default class config values"""
        default_config = {
            'THRESHOLD': 0.4,  # Similarity score threshold required for a IDTP match. 0.4 for IDEucl.
            'PRINT_CONFIG': True,  # Whether to print the config information on init. Default: False.
        }
        return default_config

    def __init__(self, config=None):
        super().__init__()
        self.fields = ['IDEucl']
        self.float_fields = self.fields
        self.summary_fields = self.fields

        # Configuration options:
        self.config = utils.init_config(config, self.get_default_config(), self.get_name())
        self.threshold = float(self.config['THRESHOLD'])


    @_timing.time
    def eval_sequence(self, data):
        """Calculates IDEucl metrics for all frames"""
        # Initialise results
        res = {'IDEucl' : 0}

        # Return result quickly if tracker or gt sequence is empty
        if data['num_tracker_dets'] == 0 or data['num_gt_dets'] == 0.:
            return res

        data['centroid'] = []
        for t, gt_det in enumerate(data['gt_dets']):
            # import pdb;pdb.set_trace()
            data['centroid'].append(self._compute_centroid(gt_det))

        oid_hid_cent = defaultdict(list)
        oid_cent = defaultdict(list)
        for t, (gt_ids_t, tracker_ids_t) in enumerate(zip(data['gt_ids'], data['tracker_ids'])):
            matches_mask = np.greater_equal(data['similarity_scores'][t], self.threshold)

            # I hope the orders of ids and boxes are maintained in `data`
            for ind, gid in enumerate(gt_ids_t):
                oid_cent[gid].append(data['centroid'][t][ind])

            match_idx_gt, match_idx_tracker = np.nonzero(matches_mask)
            for m_gid, m_tid in zip(match_idx_gt, match_idx_tracker):
                oid_hid_cent[gt_ids_t[m_gid], tracker_ids_t[m_tid]].append(data['centroid'][t][m_gid])

        oid_hid_dist = {k : np.sum(np.linalg.norm(np.diff(np.array(v), axis=0), axis=1)) for k, v in oid_hid_cent.items()}
        oid_dist = {int(k) : np.sum(np.linalg.norm(np.diff(np.array(v), axis=0), axis=1)) for k, v in oid_cent.items()}

        unique_oid = np.unique([i[0] for i in oid_hid_dist.keys()]).tolist()
        unique_hid = np.unique([i[1] for i in oid_hid_dist.keys()]).tolist()
        o_len = len(unique_oid)
        h_len = len(unique_hid)
        dist_matrix = np.zeros((o_len, h_len))
        for ((oid, hid), dist) in oid_hid_dist.items():
            oid_ind = unique_oid.index(oid)
            hid_ind = unique_hid.index(hid)
            dist_matrix[oid_ind, hid_ind] = dist

        # opt_hyp_dist contains GT ID : max dist covered by track
        opt_hyp_dist = dict.fromkeys(oid_dist.keys(), 0.)
        cost_matrix = np.max(dist_matrix) - dist_matrix
        rows, cols = linear_sum_assignment(cost_matrix)
        for (row, col) in zip(rows, cols):
            value = dist_matrix[row, col]
            opt_hyp_dist[int(unique_oid[row])] = value

        assert len(opt_hyp_dist.keys()) == len(oid_dist.keys())
        hyp_length = np.sum(list(opt_hyp_dist.values()))
        gt_length = np.sum(list(oid_dist.values()))
        id_eucl =np.mean([np.divide(a, b, out=np.zeros_like(a), where=b!=0) for a, b in zip(opt_hyp_dist.values(), oid_dist.values())])
        res['IDEucl'] = np.divide(hyp_length, gt_length, out=np.zeros_like(hyp_length), where=gt_length!=0)
        return res

    def combine_classes_class_averaged(self, all_res, ignore_empty_classes=False):
        """Combines metrics across all classes by averaging over the class values.
        If 'ignore_empty_classes' is True, then it only sums over classes with at least one gt or predicted detection.
        """
        res = {}

        for field in self.float_fields:
            if ignore_empty_classes:
                res[field] = np.mean([v[field] for v in all_res.values()
                                      if v['IDEucl'] > 0 + np.finfo('float').eps], axis=0)
            else:
                res[field] = np.mean([v[field] for v in all_res.values()], axis=0)
        return res

    def combine_classes_det_averaged(self, all_res):
        """Combines metrics across all classes by averaging over the detection values"""
        res = {}
        for field in self.float_fields:
            res[field] = self._combine_sum(all_res, field)
        res = self._compute_final_fields(res, len(all_res))
        return res

    def combine_sequences(self, all_res):
        """Combines metrics across all sequences"""
        res = {}
        for field in self.float_fields:
            res[field] = self._combine_sum(all_res, field)
        res = self._compute_final_fields(res, len(all_res))
        return res


    @staticmethod
    def _compute_centroid(box):
        box = np.array(box)
        if len(box.shape) == 1:
            centroid = (box[0:2] + box[2:4])/2
        else:
            centroid = (box[:, 0:2] + box[:, 2:4])/2
        return  np.flip(centroid, axis=1)


    @staticmethod
    def _compute_final_fields(res, res_len):
        """
        Exists only to match signature with the original Identiy class.

        """
        return {k:v/res_len for k,v in res.items()}




import numpy as np
from abc import ABC, abstractmethod
from .. import _timing
from ..utils import TrackEvalException


class _BaseMetric(ABC):
    @abstractmethod
    def __init__(self):
        self.plottable = False
        self.integer_fields = []
        self.float_fields = []
        self.array_labels = []
        self.integer_array_fields = []
        self.float_array_fields = []
        self.fields = []
        self.summary_fields = []
        self.registered = False

    #####################################################################
    # Abstract functions for subclasses to implement

    @_timing.time
    @abstractmethod
    def eval_sequence(self, data):
        ...

    @abstractmethod
    def combine_sequences(self, all_res):
        ...

    @abstractmethod
    def combine_classes_class_averaged(self, all_res, ignore_empty_classes=False):
        ...

    @ abstractmethod
    def combine_classes_det_averaged(self, all_res):
        ...

    def plot_single_tracker_results(self, all_res, tracker, output_folder, cls):
        """Plot results of metrics, only valid for metrics with self.plottable"""
        if self.plottable:
            raise NotImplementedError('plot_results is not implemented for metric %s' % self.get_name())
        else:
            pass

    #####################################################################
    # Helper functions which are useful for all metrics:

    @classmethod
    def get_name(cls):
        return cls.__name__

    @staticmethod
    def _combine_sum(all_res, field):
        """Combine sequence results via sum"""
        return sum([all_res[k][field] for k in all_res.keys()])

    @staticmethod
    def _combine_weighted_av(all_res, field, comb_res, weight_field):
        """Combine sequence results via weighted average"""
        return sum([all_res[k][field] * all_res[k][weight_field] for k in all_res.keys()]) / np.maximum(1.0, comb_res[
            weight_field])

    def print_table(self, table_res, tracker, cls):
        """Prints table of results for all sequences"""
        print('')
        metric_name = self.get_name()
        self._row_print([metric_name + ': ' + tracker + '-' + cls] + self.summary_fields)
        for seq, results in sorted(table_res.items()):
            if seq == 'COMBINED_SEQ':
                continue
            summary_res = self._summary_row(results)
            self._row_print([seq] + summary_res)
        summary_res = self._summary_row(table_res['COMBINED_SEQ'])
        self._row_print(['COMBINED'] + summary_res)

    def _summary_row(self, results_):
        vals = []
        for h in self.summary_fields:
            if h in self.float_array_fields:
                vals.append("{0:1.5g}".format(100 * np.mean(results_[h])))
            elif h in self.float_fields:
                vals.append("{0:1.5g}".format(100 * float(results_[h])))
            elif h in self.integer_fields:
                vals.append("{0:d}".format(int(results_[h])))
            else:
                raise NotImplementedError("Summary function not implemented for this field type.")
        return vals

    @staticmethod
    def _row_print(*argv):
        """Prints results in an evenly spaced rows, with more space in first row"""
        if len(argv) == 1:
            argv = argv[0]
        to_print = '%-35s' % argv[0]
        for v in argv[1:]:
            to_print += '%-10s' % str(v)
        print(to_print)

    def summary_results(self, table_res):
        """Returns a simple summary of final results for a tracker"""
        return dict(zip(self.summary_fields, self._summary_row(table_res['COMBINED_SEQ'])))

    def detailed_results(self, table_res):
        """Returns detailed final results for a tracker"""
        # Get detailed field information
        detailed_fields = self.float_fields + self.integer_fields
        for h in self.float_array_fields + self.integer_array_fields:
            for alpha in [int(100*x) for x in self.array_labels]:
                detailed_fields.append(h + '___' + str(alpha))
            detailed_fields.append(h + '___AUC')

        # Get detailed results
        detailed_results = {}
        for seq, res in table_res.items():
            detailed_row = self._detailed_row(res)
            if len(detailed_row) != len(detailed_fields):
                raise TrackEvalException(
                    'Field names and data have different sizes (%i and %i)' % (len(detailed_row), len(detailed_fields)))
            detailed_results[seq] = dict(zip(detailed_fields, detailed_row))
        return detailed_results

    def _detailed_row(self, res):
        detailed_row = []
        for h in self.float_fields + self.integer_fields:
            detailed_row.append(res[h])
        for h in self.float_array_fields + self.integer_array_fields:
            for i, alpha in enumerate([int(100 * x) for x in self.array_labels]):
                detailed_row.append(res[h][i])
            detailed_row.append(np.mean(res[h]))
        return detailed_row



import os
import json
import pycocotools.mask as cocomask
from tabulate import tabulate
from typing import Union


def _global_track_id(*, local_track_id: Union[str, int],
                     video_id: Union[str, int],
                     track_id_mapping) -> int:
    # remap local track ids into globally unique ids
    return track_id_mapping[str(video_id)][str(local_track_id)]


class GroundTruthBURSTFormatToTAOFormatConverter:
    def __init__(self, ali_format):
        self._ali_format = ali_format
        self._split = ali_format['split']
        self._categories = self._make_categories()
        self._videos = []
        self._annotations = []
        self._tracks = {}
        self._images = []
        self._next_img_id = 0
        self._next_ann_id = 0

        self._track_id_mapping = self._load_track_id_mapping()

        for seq in ali_format['sequences']:
            self._visit_seq(seq)

    def _load_track_id_mapping(self):
        id_map = {}
        next_global_track_id = 1
        for seq in self._ali_format['sequences']:
            seq_id = seq['id']
            seq_id_map = {}
            id_map[str(seq_id)] = seq_id_map
            for local_track_id in seq['track_category_ids']:
                seq_id_map[str(local_track_id)] = next_global_track_id
                next_global_track_id += 1
        return id_map

    def global_track_id(self, *, local_track_id: Union[str, int],
                        video_id: Union[str, int]) -> int:
        return _global_track_id(local_track_id=local_track_id,
                                video_id=video_id,
                                track_id_mapping=self._track_id_mapping)

    def _visit_seq(self, seq):
        self._make_video(seq)
        imgs = self._make_images(seq)
        self._make_annotations_and_tracks(seq, imgs)

    def _make_images(self, seq):
        imgs = []
        for img_path in seq['annotated_image_paths']:
            video = self._split + '/' + seq['dataset'] + '/' + seq['seq_name']
            file_name = video + '/' + img_path

            # TODO: once python 3.9 is more common, we can use this nicer and safer code
            #stripped = img_path.removesuffix('.jpg').removesuffix('.png').removeprefix('frame')
            stripped = img_path.replace('.jpg', '').replace('.png', '').replace('frame', '')

            last = stripped.split('_')[-1]
            frame_idx = int(last)

            img = {'id': self._next_img_id, 'video': video,
                   'width': seq['width'], 'height': seq['height'],
                   'file_name': file_name,
                   'frame_index': frame_idx,
                   'video_id': seq['id']}
            self._next_img_id += 1
            self._images.append(img)
            imgs.append(img)
        return imgs

    def _make_video(self, seq):
        video_id = seq['id']
        dataset = seq['dataset']
        seq_name = seq['seq_name']
        name = f'{self._split}/' + dataset + '/' + seq_name
        video = {
            'id': video_id, 'width': seq['width'], 'height': seq['height'],
            'neg_category_ids': seq['neg_category_ids'],
            'not_exhaustive_category_ids': seq['not_exhaustive_category_ids'],
            'name': name, 'metadata': {'dataset': dataset}}
        self._videos.append(video)

    def _make_annotations_and_tracks(self, seq, imgs):
        video_id = seq['id']
        segs = seq['segmentations']
        assert len(segs) == len(imgs), (len(segs), len(imgs))
        for frame_segs, img in zip(segs, imgs):
            for local_track_id, seg in frame_segs.items():
                distractors = {20, 63, 108, 180, 188, 204, 212, 247, 303, 403, 407, 415, 490, 504, 507, 513, 529, 567,
                               569, 588, 672, 691, 702, 708, 711, 720, 736, 737, 798, 813, 815, 827, 831, 851, 877, 883,
                               912, 971, 976, 1130, 1133, 1134, 1169, 1184, 1220}
                global_track_id = self.global_track_id(
                    local_track_id=local_track_id, video_id=seq['id'])
                rle = seg['rle']
                segmentation = {'counts': rle,
                                'size': [img['height'], img['width']]}
                image_id = img['id']
                category_id = int(seq['track_category_ids'][local_track_id])
                if category_id in distractors:
                    continue
                coco_bbox = cocomask.toBbox(segmentation)
                bbox = [int(x) for x in coco_bbox]
                ann = {'segmentation': segmentation, 'id': self._next_ann_id,
                       'image_id': image_id, 'category_id': category_id,
                       'track_id': global_track_id, 'video_id': video_id,
                       'bbox': bbox}
                self._next_ann_id += 1
                self._annotations.append(ann)

                if global_track_id not in self._tracks:
                    track = {'id': global_track_id, 'category_id': category_id,
                             'video_id': video_id}
                    self._tracks[global_track_id] = track

    def convert(self):
        tracks = sorted(self._tracks.values(), key=lambda t: t['id'])
        return {'videos': self._videos, 'annotations': self._annotations,
                'tracks': tracks, 'images': self._images,
                'categories': self._categories,
                'track_id_mapping': self._track_id_mapping,
                'split': self._split}

    def _make_categories(self):
        tao_categories_path = os.path.join(os.path.dirname(__file__), 'tao_categories.json')
        with open(tao_categories_path) as f:
            return json.load(f)


class PredictionBURSTFormatToTAOFormatConverter:
    def __init__(self, gt, ali_format, exemplar_guided):
        self._gt = gt
        self._ali_format = ali_format
        if 'split' in ali_format:
            self._split = ali_format['split']
            gt_split = self._gt['split']
            assert self._split == gt_split, (self._split, gt_split)
        else:
            self._split = self._gt['split']
        self._exemplar_guided = exemplar_guided
        self._result = []
        self._next_det_id = 0

        self._img_by_filename = {}
        for img in self._gt['images']:
            file_name = img['file_name']
            assert file_name not in self._img_by_filename
            self._img_by_filename[file_name] = img

        self._gt_track_by_track_id = {}
        for track in self._gt['tracks']:
            self._gt_track_by_track_id[int(track['id'])] = track

        self._filtered_out_track_ids = set()

        for seq in ali_format['sequences']:
            self._visit_seq(seq)

        if exemplar_guided and len(self._filtered_out_track_ids) > 0:
            self.print_filter_out_debug_info(ali_format)

    def print_filter_out_debug_info(self, ali_format):
        track_ids_in_pred = set()
        a_dict_for_debugging = {}
        for seq in ali_format['sequences']:
            for local_track_id in seq['track_category_ids']:
                global_track_id = _global_track_id(
                    local_track_id=local_track_id, video_id=seq['id'],
                    track_id_mapping=self._gt['track_id_mapping'])
                track_ids_in_pred.add(global_track_id)
                a_dict_for_debugging[global_track_id] = {'seq': seq,
                                                         'local_track_id': local_track_id}
        print('Number of Track ids in pred:', len(track_ids_in_pred))
        print('Exemplar Guided: Filtered out',
              len(self._filtered_out_track_ids),
              'tracks which were not found in the ground truth.')
        track_ids_after_filtering = set(d['track_id'] for d in self._result)
        print('Number of tracks after filtering:',
              len(track_ids_after_filtering))
        problem_tracks = list(
            track_ids_in_pred - track_ids_after_filtering - self._filtered_out_track_ids)
        if len(problem_tracks) > 0:
            print("\nWARNING:", len(problem_tracks),
                  "object tracks are not present. There could be a number of reasons for this:\n"
                  "(1) If you are running evaluation for the box/point exemplar-guided task then this is to be expected"
                  " because your tracker probably didn't predict masks for every ground-truth object instance.\n"
                  "(2) If you are running evaluation for the mask exemplar-guided task, then this could indicate a "
                  "problem. Assume that you copied the given first-frame object mask to your predicted result, this "
                  "should not happen. It could be that your predictions are at the wrong frame-rate i.e. you have no "
                  "predicted masks for video frames which will be evaluated.\n")

            rows = []
            for xx in problem_tracks:
                rows.append([a_dict_for_debugging[xx]['seq']['dataset'],
                             a_dict_for_debugging[xx]['seq']['seq_name'],
                             a_dict_for_debugging[xx]['local_track_id']])

            print("For your reference, the sequence name and track IDs for these missing tracks are:")
            print(tabulate(rows, ["Dataset", "Sequence Name", "Track ID"]))

    def _visit_seq(self, seq):
        dataset = seq['dataset']
        seq_name = seq['seq_name']
        assert len(seq['segmentations']) == len(seq['annotated_image_paths'])
        for frame_segs, img_path in zip(seq['segmentations'],
                                        seq['annotated_image_paths']):
            for local_track_id_str, track_det in frame_segs.items():
                rle = track_det['rle']

                file_name = self._split + '/' + dataset + '/' + seq_name + '/' + img_path
                # the result might have a higher frame rate than the ground truth
                if file_name not in self._img_by_filename:
                    continue

                img = self._img_by_filename[file_name]
                img_id = img['id']
                height = img['height']
                width = img['width']
                segmentation = {'counts': rle, 'size': [height, width]}

                local_track_id = int(local_track_id_str)
                if self._exemplar_guided:
                    global_track_id = _global_track_id(
                        local_track_id=local_track_id, video_id=seq['id'],
                        track_id_mapping=self._gt['track_id_mapping'])
                else:
                    global_track_id = local_track_id
                coco_bbox = cocomask.toBbox(segmentation)
                bbox = [int(x) for x in coco_bbox]
                det = {'id': self._next_det_id, 'image_id': img_id,
                       'track_id': global_track_id, 'bbox': bbox,
                       'segmentation': segmentation}
                if self._exemplar_guided:
                    if global_track_id not in self._gt_track_by_track_id:
                        self._filtered_out_track_ids.add(global_track_id)
                        continue
                    gt_track = self._gt_track_by_track_id[global_track_id]
                    category_id = gt_track['category_id']
                    det['category_id'] = category_id
                elif 'category_id' in track_det:
                    det['category_id'] = track_det['category_id']
                else:
                    category_id = seq['track_category_ids'][local_track_id_str]
                    det['category_id'] = category_id
                self._next_det_id += 1
                if 'score' in track_det:
                    det['score'] = track_det['score']
                else:
                    det['score'] = 1.0
                self._result.append(det)

    def convert(self):
        return self._result



import json
import argparse
from .format_converter import GroundTruthBURSTFormatToTAOFormatConverter, PredictionBURSTFormatToTAOFormatConverter


def main(args):
    with open(args.gt_input_file) as f:
        ali_format_gt = json.load(f)
    tao_format_gt = GroundTruthBURSTFormatToTAOFormatConverter(
        ali_format_gt, args.split).convert()
    with open(args.gt_output_file, 'w') as f:
        json.dump(tao_format_gt, f)

    if args.pred_input_file is None:
        return
    with open(args.pred_input_file) as f:
        ali_format_pred = json.load(f)
    tao_format_pred = PredictionBURSTFormatToTAOFormatConverter(
        tao_format_gt, ali_format_pred, args.split,
        args.exemplar_guided).convert()
    with open(args.pred_output_file, 'w') as f:
        json.dump(tao_format_pred, f)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--gt_input_file', type=str,
        default='../data/gt/tsunami/exemplar_guided/validation_all_annotations.json')
    parser.add_argument('--gt_output_file', type=str,
                        default='/tmp/val_gt.json')
    parser.add_argument('--pred_input_file', type=str,
                        default='../data/trackers/tsunami/exemplar_guided/STCN_off_the_shelf/data/results.json')
    parser.add_argument('--pred_output_file', type=str,
                        default='/tmp/pred.json')
    parser.add_argument('--split', type=str, default='validation')
    parser.add_argument('--exemplar_guided', type=bool, default=True)
    args_ = parser.parse_args()
    main(args_)



"""
Thresholder

Author: Jonathon Luiten

Simply reads in a set of detection, thresholds them at a certain score threshold, and writes them out again.
"""

import os
import sys
from multiprocessing.pool import Pool
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))
from trackeval.baselines import baseline_utils as butils
from trackeval.utils import get_code_path

THRESHOLD = 0.2

code_path = get_code_path()
config = {
    'INPUT_FOL': os.path.join(code_path, 'data/detections/rob_mots/{split}/non_overlap_supplied/data/'),
    'OUTPUT_FOL': os.path.join(code_path, 'data/detections/rob_mots/{split}/threshold_' + str(100*THRESHOLD) + '/data/'),
    'SPLIT': 'train',  # valid: 'train', 'val', 'test'.
    'Benchmarks': None,  # If None, all benchmarks in SPLIT.

    'Num_Parallel_Cores': None,  # If None, run without parallel.

    'DETECTION_THRESHOLD': THRESHOLD,
}


def do_sequence(seq_file):

    # Load input data from file (e.g. provided detections)
    # data format: data['cls'][t] = {'ids', 'scores', 'im_hs', 'im_ws', 'mask_rles'}
    data = butils.load_seq(seq_file)

    # Where to accumulate output data for writing out
    output_data = []

    # Run for each class.
    for cls, cls_data in data.items():

        # Run for each timestep.
        for timestep, t_data in enumerate(cls_data):

            # Threshold detections.
            t_data = butils.threshold(t_data, config['DETECTION_THRESHOLD'])

            # Save result in output format to write to file later.
            # Output Format = [timestep ID class score im_h im_w mask_RLE]
            for i in range(len(t_data['ids'])):
                row = [timestep, int(t_data['ids'][i]), cls, t_data['scores'][i], t_data['im_hs'][i],
                       t_data['im_ws'][i], t_data['mask_rles'][i]]
                output_data.append(row)

    # Write results to file
    out_file = seq_file.replace(config['INPUT_FOL'].format(split=config['SPLIT']),
                                config['OUTPUT_FOL'].format(split=config['SPLIT']))
    butils.write_seq(output_data, out_file)

    print('DONE:', seq_todo)


if __name__ == '__main__':

    # Required to fix bug in multiprocessing on windows.
    freeze_support()

    # Obtain list of sequences to run tracker for.
    if config['Benchmarks']:
        benchmarks = config['Benchmarks']
    else:
        benchmarks = ['davis_unsupervised', 'kitti_mots', 'youtube_vis', 'ovis', 'bdd_mots', 'tao']
        if config['SPLIT'] != 'train':
            benchmarks += ['waymo', 'mots_challenge']
    seqs_todo = []
    for bench in benchmarks:
        bench_fol = os.path.join(config['INPUT_FOL'].format(split=config['SPLIT']), bench)
        seqs_todo += [os.path.join(bench_fol, seq) for seq in os.listdir(bench_fol)]

    # Run in parallel
    if config['Num_Parallel_Cores']:
        with Pool(config['Num_Parallel_Cores']) as pool:
            results = pool.map(do_sequence, seqs_todo)

    # Run in series
    else:
        for seq_todo in seqs_todo:
            do_sequence(seq_todo)




import os
from .burst_helpers.burst_base import BURSTBase
from .burst_helpers.format_converter import GroundTruthBURSTFormatToTAOFormatConverter, PredictionBURSTFormatToTAOFormatConverter
from .. import utils


class BURST(BURSTBase):
    """Dataset class for TAO tracking"""

    @staticmethod
    def get_default_dataset_config():
        tao_config = BURSTBase.get_default_dataset_config()
        code_path = utils.get_code_path()

        # e.g. 'data/gt/tsunami/exemplar_guided/'
        tao_config['GT_FOLDER'] = os.path.join(
            code_path, 'data/gt/burst/val/')  # Location of GT data
        # e.g. 'data/trackers/tsunami/exemplar_guided/mask_guided/validation/'
        tao_config['TRACKERS_FOLDER'] = os.path.join(
            code_path, 'data/trackers/burst/class-guided/')  # Trackers location
        # set to True or False
        tao_config['EXEMPLAR_GUIDED'] = False
        return tao_config

    def _iou_type(self):
        return 'mask'

    def _box_or_mask_from_det(self, det):
        return det['segmentation']

    def _calculate_area_for_ann(self, ann):
        import pycocotools.mask as cocomask
        return cocomask.area(ann["segmentation"])

    def _calculate_similarities(self, gt_dets_t, tracker_dets_t):
        similarity_scores = self._calculate_mask_ious(gt_dets_t, tracker_dets_t, is_encoded=True, do_ioa=False)
        return similarity_scores

    def _is_exemplar_guided(self):
        exemplar_guided = self.config['EXEMPLAR_GUIDED']
        return exemplar_guided

    def _postproc_ground_truth_data(self, data):
        return GroundTruthBURSTFormatToTAOFormatConverter(data).convert()

    def _postproc_prediction_data(self, data):
        return PredictionBURSTFormatToTAOFormatConverter(
            self.gt_data, data,
            exemplar_guided=self._is_exemplar_guided()).convert()



"""."""

from typing import Dict, Sequence, Union
import argparse


def modify_with_cmd_line(config: Dict[str, Union[str, int, float, bool]],
                         args: Union[None, Sequence[str]]) -> None:
    parser = argparse.ArgumentParser()
    for setting in config.keys():
        if isinstance(config[setting], list) or config[setting] is None:
            parser.add_argument("--" + setting, nargs='+')
        else:
            parser.add_argument("--" + setting)
    namespace = parser.parse_args(args).__dict__

    for setting in namespace.keys():
        if namespace[setting] is not None:
            if type(config[setting]) == type(True):
                if namespace[setting] == 'True':
                    x = True
                elif namespace[setting] == 'False':
                    x = False
                else:
                    raise Exception('Command line parameter ' + setting + 'must be True or False')
            elif type(config[setting]) == type(1):
                x = int(namespace[setting])
            elif type(namespace[setting]) == type(None):
                x = None
            else:
                x = namespace[setting]
            config[setting] = x



import os
import json
import numpy as np
from scipy.optimize import linear_sum_assignment
from ..utils import TrackEvalException
from ._base_dataset import _BaseDataset
from .. import utils
from .. import _timing


class BDD100K(_BaseDataset):
    """Dataset class for BDD100K tracking"""

    @staticmethod
    def get_default_dataset_config():
        """Default class config values"""
        code_path = utils.get_code_path()
        default_config = {
            'GT_FOLDER': os.path.join(code_path, 'data/gt/bdd100k/bdd100k_val'),  # Location of GT data
            'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/bdd100k/bdd100k_val'),  # Trackers location
            'OUTPUT_FOLDER': None,  # Where to save eval results (if None, same as TRACKERS_FOLDER)
            'TRACKERS_TO_EVAL': None,  # Filenames of trackers to eval (if None, all in folder)
            'CLASSES_TO_EVAL': ['pedestrian', 'rider', 'car', 'bus', 'truck', 'train', 'motorcycle', 'bicycle'],
            # Valid: ['pedestrian', 'rider', 'car', 'bus', 'truck', 'train', 'motorcycle', 'bicycle']
            'SPLIT_TO_EVAL': 'val',  # Valid: 'training', 'val',
            'INPUT_AS_ZIP': False,  # Whether tracker input files are zipped
            'PRINT_CONFIG': True,  # Whether to print current config
            'TRACKER_SUB_FOLDER': 'data',  # Tracker files are in TRACKER_FOLDER/tracker_name/TRACKER_SUB_FOLDER
            'OUTPUT_SUB_FOLDER': '',  # Output files are saved in OUTPUT_FOLDER/tracker_name/OUTPUT_SUB_FOLDER
            'TRACKER_DISPLAY_NAMES': None,  # Names of trackers to display, if None: TRACKERS_TO_EVAL
        }
        return default_config

    def __init__(self, config=None):
        """Initialise dataset, checking that all required files are present"""
        super().__init__()
        # Fill non-given config values with defaults
        self.config = utils.init_config(config, self.get_default_dataset_config(), self.get_name())
        self.gt_fol = self.config['GT_FOLDER']
        self.tracker_fol = self.config['TRACKERS_FOLDER']
        self.should_classes_combine = True
        self.use_super_categories = True

        self.output_fol = self.config['OUTPUT_FOLDER']
        if self.output_fol is None:
            self.output_fol = self.tracker_fol

        self.tracker_sub_fol = self.config['TRACKER_SUB_FOLDER']
        self.output_sub_fol = self.config['OUTPUT_SUB_FOLDER']

        # Get classes to eval
        self.valid_classes = ['pedestrian', 'rider', 'car', 'bus', 'truck', 'train', 'motorcycle', 'bicycle']
        self.class_list = [cls.lower() if cls.lower() in self.valid_classes else None
                           for cls in self.config['CLASSES_TO_EVAL']]
        if not all(self.class_list):
            raise TrackEvalException('Attempted to evaluate an invalid class. Only classes [pedestrian, rider, car, '
                                     'bus, truck, train, motorcycle, bicycle] are valid.')
        self.super_categories = {"HUMAN": [cls for cls in ["pedestrian", "rider"] if cls in self.class_list],
                                 "VEHICLE": [cls for cls in ["car", "truck", "bus", "train"] if cls in self.class_list],
                                 "BIKE": [cls for cls in ["motorcycle", "bicycle"] if cls in self.class_list]}
        self.distractor_classes = ['other person', 'trailer', 'other vehicle']
        self.class_name_to_class_id = {'pedestrian': 1, 'rider': 2, 'other person': 3, 'car': 4, 'bus': 5, 'truck': 6,
                                       'train': 7, 'trailer': 8, 'other vehicle': 9, 'motorcycle': 10, 'bicycle': 11}

        # Get sequences to eval
        self.seq_list = []
        self.seq_lengths = {}

        self.seq_list = [seq_file.replace('.json', '') for seq_file in os.listdir(self.gt_fol)]

        # Get trackers to eval
        if self.config['TRACKERS_TO_EVAL'] is None:
            self.tracker_list = os.listdir(self.tracker_fol)
        else:
            self.tracker_list = self.config['TRACKERS_TO_EVAL']

        if self.config['TRACKER_DISPLAY_NAMES'] is None:
            self.tracker_to_disp = dict(zip(self.tracker_list, self.tracker_list))
        elif (self.config['TRACKERS_TO_EVAL'] is not None) and (
                len(self.config['TRACKER_DISPLAY_NAMES']) == len(self.tracker_list)):
            self.tracker_to_disp = dict(zip(self.tracker_list, self.config['TRACKER_DISPLAY_NAMES']))
        else:
            raise TrackEvalException('List of tracker files and tracker display names do not match.')

        for tracker in self.tracker_list:
            for seq in self.seq_list:
                curr_file = os.path.join(self.tracker_fol, tracker, self.tracker_sub_fol, seq + '.json')
                if not os.path.isfile(curr_file):
                    print('Tracker file not found: ' + curr_file)
                    raise TrackEvalException(
                        'Tracker file not found: ' + tracker + '/' + self.tracker_sub_fol + '/' + os.path.basename(
                            curr_file))

    def get_display_name(self, tracker):
        return self.tracker_to_disp[tracker]

    def _load_raw_file(self, tracker, seq, is_gt):
        """Load a file (gt or tracker) in the BDD100K format

        If is_gt, this returns a dict which contains the fields:
        [gt_ids, gt_classes] : list (for each timestep) of 1D NDArrays (for each det).
        [gt_dets, gt_crowd_ignore_regions]: list (for each timestep) of lists of detections.

        if not is_gt, this returns a dict which contains the fields:
        [tracker_ids, tracker_classes, tracker_confidences] : list (for each timestep) of 1D NDArrays (for each det).
        [tracker_dets]: list (for each timestep) of lists of detections.
        """
        # File location
        if is_gt:
            file = os.path.join(self.gt_fol, seq + '.json')
        else:
            file = os.path.join(self.tracker_fol, tracker, self.tracker_sub_fol, seq + '.json')

        with open(file) as f:
            data = json.load(f)

        # sort data by frame index
        data = sorted(data, key=lambda x: x['index'])

        # check sequence length
        if is_gt:
            self.seq_lengths[seq] = len(data)
            num_timesteps = len(data)
        else:
            num_timesteps = self.seq_lengths[seq]
            if num_timesteps != len(data):
                raise TrackEvalException('Number of ground truth and tracker timesteps do not match for sequence %s'
                                         % seq)

        # Convert data to required format
        data_keys = ['ids', 'classes', 'dets']
        if is_gt:
            data_keys += ['gt_crowd_ignore_regions']
        raw_data = {key: [None] * num_timesteps for key in data_keys}
        for t in range(num_timesteps):
            ig_ids = []
            keep_ids = []
            for i in range(len(data[t]['labels'])):
                ann = data[t]['labels'][i]
                if is_gt and (ann['category'] in self.distractor_classes or 'attributes' in ann.keys()
                              and ann['attributes']['Crowd']):
                    ig_ids.append(i)
                else:
                    keep_ids.append(i)

            if keep_ids:
                raw_data['dets'][t] = np.atleast_2d([[data[t]['labels'][i]['box2d']['x1'],
                                                      data[t]['labels'][i]['box2d']['y1'],
                                                      data[t]['labels'][i]['box2d']['x2'],
                                                      data[t]['labels'][i]['box2d']['y2']
                                                      ] for i in keep_ids]).astype(float)
                raw_data['ids'][t] = np.atleast_1d([data[t]['labels'][i]['id'] for i in keep_ids]).astype(int)
                raw_data['classes'][t] = np.atleast_1d([self.class_name_to_class_id[data[t]['labels'][i]['category']]
                                                        for i in keep_ids]).astype(int)
            else:
                raw_data['dets'][t] = np.empty((0, 4)).astype(float)
                raw_data['ids'][t] = np.empty(0).astype(int)
                raw_data['classes'][t] = np.empty(0).astype(int)

            if is_gt:
                if ig_ids:
                    raw_data['gt_crowd_ignore_regions'][t] = np.atleast_2d([[data[t]['labels'][i]['box2d']['x1'],
                                                                             data[t]['labels'][i]['box2d']['y1'],
                                                                             data[t]['labels'][i]['box2d']['x2'],
                                                                             data[t]['labels'][i]['box2d']['y2']
                                                                             ] for i in ig_ids]).astype(float)
                else:
                    raw_data['gt_crowd_ignore_regions'][t] = np.empty((0, 4)).astype(float)

        if is_gt:
            key_map = {'ids': 'gt_ids',
                       'classes': 'gt_classes',
                       'dets': 'gt_dets'}
        else:
            key_map = {'ids': 'tracker_ids',
                       'classes': 'tracker_classes',
                       'dets': 'tracker_dets'}
        for k, v in key_map.items():
            raw_data[v] = raw_data.pop(k)
        raw_data['num_timesteps'] = num_timesteps
        return raw_data

    @_timing.time
    def get_preprocessed_seq_data(self, raw_data, cls):
        """ Preprocess data for a single sequence for a single class ready for evaluation.
        Inputs:
             - raw_data is a dict containing the data for the sequence already read in by get_raw_seq_data().
             - cls is the class to be evaluated.
        Outputs:
             - data is a dict containing all of the information that metrics need to perform evaluation.
                It contains the following fields:
                    [num_timesteps, num_gt_ids, num_tracker_ids, num_gt_dets, num_tracker_dets] : integers.
                    [gt_ids, tracker_ids, tracker_confidences]: list (for each timestep) of 1D NDArrays (for each det).
                    [gt_dets, tracker_dets]: list (for each timestep) of lists of detections.
                    [similarity_scores]: list (for each timestep) of 2D NDArrays.
        Notes:
            General preprocessing (preproc) occurs in 4 steps. Some datasets may not use all of these steps.
                1) Extract only detections relevant for the class to be evaluated (including distractor detections).
                2) Match gt dets and tracker dets. Remove tracker dets that are matched to a gt det that is of a
                    distractor class, or otherwise marked as to be removed.
                3) Remove unmatched tracker dets if they fall within a crowd ignore region or don't meet a certain
                    other criteria (e.g. are too small).
                4) Remove gt dets that were only useful for preprocessing and not for actual evaluation.
            After the above preprocessing steps, this function also calculates the number of gt and tracker detections
                and unique track ids. It also relabels gt and tracker ids to be contiguous and checks that ids are
                unique within each timestep.

        BDD100K:
            In BDD100K, the 4 preproc steps are as follow:
                1) There are eight classes (pedestrian, rider, car, bus, truck, train, motorcycle, bicycle)
                    which are evaluated separately.
                2) For BDD100K there is no removal of matched tracker dets.
                3) Crowd ignore regions are used to remove unmatched detections.
                4) No removal of gt dets.
        """
        cls_id = self.class_name_to_class_id[cls]

        data_keys = ['gt_ids', 'tracker_ids', 'gt_dets', 'tracker_dets', 'similarity_scores']
        data = {key: [None] * raw_data['num_timesteps'] for key in data_keys}
        unique_gt_ids = []
        unique_tracker_ids = []
        num_gt_dets = 0
        num_tracker_dets = 0
        for t in range(raw_data['num_timesteps']):

            # Only extract relevant dets for this class for preproc and eval (cls)
            gt_class_mask = np.atleast_1d(raw_data['gt_classes'][t] == cls_id)
            gt_class_mask = gt_class_mask.astype(bool)
            gt_ids = raw_data['gt_ids'][t][gt_class_mask]
            gt_dets = raw_data['gt_dets'][t][gt_class_mask]

            tracker_class_mask = np.atleast_1d(raw_data['tracker_classes'][t] == cls_id)
            tracker_class_mask = tracker_class_mask.astype(bool)
            tracker_ids = raw_data['tracker_ids'][t][tracker_class_mask]
            tracker_dets = raw_data['tracker_dets'][t][tracker_class_mask]
            similarity_scores = raw_data['similarity_scores'][t][gt_class_mask, :][:, tracker_class_mask]

            # Match tracker and gt dets (with hungarian algorithm)
            unmatched_indices = np.arange(tracker_ids.shape[0])
            if gt_ids.shape[0] > 0 and tracker_ids.shape[0] > 0:
                matching_scores = similarity_scores.copy()
                matching_scores[matching_scores < 0.5 - np.finfo('float').eps] = 0
                match_rows, match_cols = linear_sum_assignment(-matching_scores)
                actually_matched_mask = matching_scores[match_rows, match_cols] > 0 + np.finfo('float').eps
                match_cols = match_cols[actually_matched_mask]
                unmatched_indices = np.delete(unmatched_indices, match_cols, axis=0)

            # For unmatched tracker dets, remove those that are greater than 50% within a crowd ignore region.
            unmatched_tracker_dets = tracker_dets[unmatched_indices, :]
            crowd_ignore_regions = raw_data['gt_crowd_ignore_regions'][t]
            intersection_with_ignore_region = self._calculate_box_ious(unmatched_tracker_dets, crowd_ignore_regions,
                                                                       box_format='x0y0x1y1', do_ioa=True)
            is_within_crowd_ignore_region = np.any(intersection_with_ignore_region > 0.5 + np.finfo('float').eps,
                                                   axis=1)

            # Apply preprocessing to remove unwanted tracker dets.
            to_remove_tracker = unmatched_indices[is_within_crowd_ignore_region]
            data['tracker_ids'][t] = np.delete(tracker_ids, to_remove_tracker, axis=0)
            data['tracker_dets'][t] = np.delete(tracker_dets, to_remove_tracker, axis=0)
            similarity_scores = np.delete(similarity_scores, to_remove_tracker, axis=1)

            data['gt_ids'][t] = gt_ids
            data['gt_dets'][t] = gt_dets
            data['similarity_scores'][t] = similarity_scores

            unique_gt_ids += list(np.unique(data['gt_ids'][t]))
            unique_tracker_ids += list(np.unique(data['tracker_ids'][t]))
            num_tracker_dets += len(data['tracker_ids'][t])
            num_gt_dets += len(data['gt_ids'][t])

        # Re-label IDs such that there are no empty IDs
        if len(unique_gt_ids) > 0:
            unique_gt_ids = np.unique(unique_gt_ids)
            gt_id_map = np.nan * np.ones((np.max(unique_gt_ids) + 1))
            gt_id_map[unique_gt_ids] = np.arange(len(unique_gt_ids))
            for t in range(raw_data['num_timesteps']):
                if len(data['gt_ids'][t]) > 0:
                    data['gt_ids'][t] = gt_id_map[data['gt_ids'][t]].astype(int)
        if len(unique_tracker_ids) > 0:
            unique_tracker_ids = np.unique(unique_tracker_ids)
            tracker_id_map = np.nan * np.ones((np.max(unique_tracker_ids) + 1))
            tracker_id_map[unique_tracker_ids] = np.arange(len(unique_tracker_ids))
            for t in range(raw_data['num_timesteps']):
                if len(data['tracker_ids'][t]) > 0:
                    data['tracker_ids'][t] = tracker_id_map[data['tracker_ids'][t]].astype(int)

        # Record overview statistics.
        data['num_tracker_dets'] = num_tracker_dets
        data['num_gt_dets'] = num_gt_dets
        data['num_tracker_ids'] = len(unique_tracker_ids)
        data['num_gt_ids'] = len(unique_gt_ids)
        data['num_timesteps'] = raw_data['num_timesteps']

        # Ensure that ids are unique per timestep.
        self._check_unique_ids(data)

        return data

    def _calculate_similarities(self, gt_dets_t, tracker_dets_t):
        similarity_scores = self._calculate_box_ious(gt_dets_t, tracker_dets_t, box_format='x0y0x1y1')
        return similarity_scores
